/* Copyright (c) 1996 University of Utah.  All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for NON-COMMERCIAL purposes and without fee is
 * hereby granted provided that this copyright notice appears in all
 * copies.
 *
 * See http://www.cs.utah.edu/~hamlet for more educational resources.
 *
 * The University of Utah MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT
 * THE SUITABILITY OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED,
 * INCLUDING BUT NOT LIMITED TO THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, OR
 * NON-INFRINGEMENT. The University of Utah SHALL NOT BE LIABLE FOR
 * ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
 * DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
 */

package Robot;

import java.awt.*;
import hamlet.*;
import Robot.*;

//
//  Class robotPanel (containing one (1) robot)
//
public class robotPanel extends Panel implements Runnable {
	
    public static int xOffset = 150;
    public static int yOffset = 10;
    public static double g = 9.8;		/*  Gravity! */
    
    public robotControls controlPanel;
    public Thread animatorThread;
    
    //  these variables describe the robot as it appears onscreen
    public double torsoLength, thighLength, shinLength;
    int windowHeight;
    public Point shoulderPos, kneePos, hipPos;
    public double ankleAngle, hipAngle, kneeAngle;
    public double kneeX, kneeY, hipX, hipY, shoulderY;
    int destAngle = 180;
    public double alpha, beta, gamma;

    int statedKneeY, statedHipY, statedShoulderY;
    int statedKneeX, statedHipX;

    //  these variables are for use of the the labeling system
    public int statedTorso;
    public NumLabel shoulderLabelY, hipLabelY, kneeLabelY;
    public NumLabel hipLabelX, kneeLabelX;
    
    // double buffering scheme
    Image offImage;
    Graphics offGraphics;

    
public robotPanel()  {
    shoulderPos = new Point(0,0);  
    kneePos = new Point(0,0);  
    hipPos = new Point(0,0);
    animatorThread = null;
    statedTorso = 300;
    kneeAngle = 60;
    setVars();
}
      
      
public void reset() {
    halt();
    offImage = createImage( bounds().width, bounds().height );
    offGraphics = offImage.getGraphics();
    statedTorso = controlPanel.torsoField.getIntText();
    kneeAngle = controlPanel.kneeField.getIntText();
    /*statedTorso = retrieveText( controlPanel.torsoField, 0, 700 );
    kneeAngle = retrieveText( controlPanel.kneeField, 0, 180 );*/
    setVars();
    update( getGraphics() );
}
 
    
public void halt() {
    animatorThread = null;
    validate();

    controlPanel.torsoField.setIntText(statedTorso);
    controlPanel.kneeField.setIntText((int)Math.round(kneeAngle));
}
 
    
public synchronized void update( Graphics g )  {
    int radius1 = 30,     	//  radius for the smaller blue arcs
	radius2 = 60;		//  radius for the larger green arcs
    offGraphics.setColor( Color.gray );
    offGraphics.fillRect(  0,0,bounds().width, bounds().height );
    offGraphics.setColor( Color.white );
    offGraphics.clearRect( 10,10,bounds().width-20, bounds().height-20 );
    
    //  draw the light grey axis lines
    offGraphics.setColor( Color.gray.brighter() );
    offGraphics.drawLine( 0, bounds().height-yOffset-hipPos.y, 
			  bounds().width, bounds().height-yOffset-hipPos.y );
    offGraphics.drawLine( 0, bounds().height-yOffset-kneePos.y, 
			  bounds().width, bounds().height-yOffset-kneePos.y );
    offGraphics.drawLine( 0, bounds().height-yOffset-shoulderPos.y, 
			  bounds().width, 
			  bounds().height-yOffset-shoulderPos.y );
    offGraphics.drawLine( xOffset, 0, xOffset, bounds().height-yOffset );
    offGraphics.drawLine( xOffset+hipPos.x, 0, 
			  xOffset+hipPos.x, bounds().height-yOffset );
    offGraphics.drawLine( xOffset+kneePos.x, 0, 
			  xOffset+kneePos.x, bounds().height-yOffset );
    
    //  draw number labels on them
  	offGraphics.setColor( Color.black );
  	offGraphics.setFont( getFont() );
	shoulderLabelY  = new NumLabel( this, statedShoulderY, "mm", 
	      	20, (bounds().height-yOffset-shoulderPos.y-12), Label.LEFT );
	hipLabelY = new NumLabel( this, statedHipY, "mm", 
		20, bounds().height-yOffset-hipPos.y-12, Label.LEFT );

 	hipLabelX = new NumLabel(this, statedHipX, "mm",
		xOffset+hipPos.x-1, 20, Label.RIGHT);

	kneeLabelY = new NumLabel( this, statedKneeY, "mm", 
		20, bounds().height-yOffset-kneePos.y-12, Label.LEFT );

	kneeLabelX = new NumLabel(this, statedKneeX, "mm",
		xOffset+kneePos.x+1, 20, Label.LEFT);

	shoulderLabelY.paint( offGraphics );
	hipLabelY.paint( offGraphics );
	hipLabelX.paint( offGraphics );
	kneeLabelY.paint( offGraphics );
	kneeLabelX.paint( offGraphics );
	
    //  draw angle arcs in pretty colors, and labels in matching colors
    offGraphics.setColor( Color.blue.darker() );
    offGraphics.drawString( ""+(int)Math.round(ankleAngle), xOffset+20, 
			   bounds().height-yOffset-2 );
    offGraphics.drawArc( xOffset-(radius1/2), 
			bounds().height-yOffset-(radius1/2), 
			radius1, radius1,  
			 0, (int) Math.round(ankleAngle));			//  ankleAngle
    offGraphics.drawString( ""+(int)Math.round(beta), 
			   xOffset+kneePos.x, 
			   bounds().height-yOffset-kneePos.y-2 );
    offGraphics.drawArc( xOffset+kneePos.x-(radius1/2), 
			bounds().height-yOffset-kneePos.y-(radius1/2), 
			 radius1, radius1, 180-(int)Math.round(beta), (int)Math.round(beta) ); 	//  beta
    offGraphics.drawString( ""+(int)Math.round(gamma), 
			   xOffset+hipPos.x-20, 
			   bounds().height-yOffset-hipPos.y-2 );
    offGraphics.drawArc( xOffset+hipPos.x-(radius1/2), 
			 bounds().height-yOffset-hipPos.y-(radius1/2), 
			 radius1, radius1, 0, (int)Math.round(gamma) );	//  gamma
    
    offGraphics.setColor( Color.green.darker().darker().darker() );
    offGraphics.drawString( ""+(int)Math.round(kneeAngle), 
			    xOffset+kneePos.x-(radius2-10), 
			    bounds().height-yOffset-kneePos.y-2 );
    offGraphics.drawArc( xOffset+kneePos.x-(radius2/2), 
			 bounds().height-yOffset-kneePos.y-(radius2/2), 	
			 radius2, radius2,180-(int)Math.round(beta), 
			 (int) Math.round(kneeAngle));				//  kneeAngle
    offGraphics.drawString( ""+(int)Math.round(hipAngle), 
			    xOffset+hipPos.x+(radius2/2)+5, 
			    bounds().height-yOffset-hipPos.y-5 );
    offGraphics.drawArc( xOffset+hipPos.x-(radius2/2), 
			 bounds().height-yOffset-hipPos.y-(radius2/2), 
			 radius2, radius2,(int)Math.round(gamma), 
			 -(int) Math.round(hipAngle) );				//  hipAngle
    
    //  lowest triangle
    offGraphics.setColor( Color.pink.darker() );	//  draw the leg in pink
    offGraphics.drawLine( xOffset, bounds().height-yOffset, 
			  xOffset+kneePos.x, bounds().height-yOffset-kneePos.y );
    offGraphics.setColor( Color.gray.darker() );
    offGraphics.drawLine( xOffset+kneePos.x, bounds().height-yOffset-kneePos.y, 
			  xOffset+kneePos.x, bounds().height-yOffset );
    offGraphics.drawLine( xOffset, bounds().height-yOffset, 
			  xOffset+kneePos.x, bounds().height-yOffset );
    
    //  middle triangle
    offGraphics.drawLine( xOffset+kneePos.x, bounds().height-yOffset-kneePos.y,
			  xOffset+hipPos.x, bounds().height-yOffset-kneePos.y );
    offGraphics.drawLine( xOffset+hipPos.x, bounds().height-yOffset-hipPos.y,
			  xOffset+hipPos.x, bounds().height-yOffset-kneePos.y );
    offGraphics.setColor( Color.pink.darker() );	//  draw the leg in pink
    offGraphics.drawLine( xOffset+kneePos.x, bounds().height-yOffset-kneePos.y,
			  xOffset+hipPos.x, bounds().height-yOffset-hipPos.y );
    
    //  torso triangle
    offGraphics.drawLine( xOffset+hipPos.x, bounds().height-yOffset-hipPos.y,
			  xOffset, bounds().height-yOffset-shoulderPos.y );
    offGraphics.setColor( Color.gray.darker() );
    offGraphics.drawLine( xOffset+hipPos.x, bounds().height-yOffset-hipPos.y,
			  xOffset, bounds().height-yOffset-hipPos.y );
    offGraphics.drawLine( xOffset, bounds().height-yOffset-hipPos.y, 
			  xOffset, bounds().height-yOffset-shoulderPos.y );
    
    // puke it to the screen
    g.drawImage( offImage, 0, 0, this );
} 
    
    
public synchronized void paint ( Graphics g )  {
    offImage = createImage( bounds().width, bounds().height );
    offGraphics = offImage.getGraphics();
    update( g );
}


public synchronized void go() {
    animatorThread = new Thread( this );
    offImage = createImage( bounds().width, bounds().height );
    offGraphics = offImage.getGraphics();
    
    //  get the user input from the control panel
    //  the torso value the user gives is only used for labeling purposes
    statedTorso = controlPanel.torsoField.getIntText();
    destAngle = controlPanel.kneeField.getIntText();
    setVars();
    animatorThread.start();
} 
    
 
public void run()  {
    // animatorThread.setPriority(Thread.MIN_PRIORITY);
    long startTime = System.currentTimeMillis();
     
     //  let's animate
    if (kneeAngle == destAngle) {
	setVars();
	update( getGraphics() );
	getToolkit().sync();
	return;
    }	

     while ( (Thread.currentThread() == animatorThread) && 
	                                      ( animatorThread != null)
	     && (kneeAngle!=destAngle) ) {
	 		if (kneeAngle>destAngle)
	     		kneeAngle -= 1;
	 		else kneeAngle += 1;
	 		setVars();
	 		update( getGraphics() );
   			getToolkit().sync();
			try {
	     		startTime += 20;
	     		Thread.sleep(
			   Math.max(0, startTime-System.currentTimeMillis()));
	 		}
	 		catch (InterruptedException e)  {
	     		break;
	 		}
     	}
}


public void setVars() {

	//  sets the coordinates of the robot that actually appears onscreen
    torsoLength = (int)Math.round( (double)(windowHeight-35)/2.26 );
      // System.out.println("In setVars(), our actual torsoLength is "
      // +torsoLength );
    alpha = ankleAngle = kneeAngle/2;
    beta = kneeAngle - alpha;
    thighLength = .67*torsoLength;
    shinLength = .59*torsoLength;
    
    kneeX = shinLength * Math.cos( radians(alpha));
    kneePos.x = (int)Math.round(kneeX);

    kneeY = shinLength * Math.sin( radians(alpha));
    kneePos.y = (int)Math.round(kneeY);

    hipX = kneeX - thighLength * Math.cos( radians(beta) );
    hipPos.x = (int)Math.round(hipX);

    hipY = kneeY + thighLength * Math.sin( radians(beta) );
    hipPos.y = (int)Math.round(hipY);
    
    gamma = degrees( Math.acos( -(double)hipPos.x/(double)torsoLength ) );

    hipAngle = gamma + beta;

    shoulderY = hipY + torsoLength * Math.sin( radians(gamma) );
    shoulderPos.y = (int) Math.round(shoulderY);
    shoulderPos.x = 0;
    // printOut();
    
    //  set the variables for the labeling system 
    
    double x = .59 * statedTorso * Math.cos(radians(alpha));
    double y = .59 * statedTorso * Math.sin(radians(alpha));

    statedKneeX = (int)Math.round(x);
    statedKneeY = (int)Math.round(y);

    x = x - .67*statedTorso * Math.cos(radians(beta));
    y = y + .67 * statedTorso * Math.sin(radians(beta));

    statedHipX = (int)Math.round(x);
    statedHipY = (int)Math.round(y);

    y = y + statedTorso * Math.sin(radians(gamma));

    statedShoulderY = (int)Math.round(y);
}
    
    
synchronized int retrieveText( TextField tf, int min, int max ) {
    String  s;
    int average, value;
    
    s = tf.getText();
    average = (int)Math.round( (max-min)/2 );
    try {
	value = Integer.parseInt(s);
    } catch (Exception e) {
	value = average;
    }
    if ( value > max ) value = max;
    if ( value < min ) value = min;
    tf.setText( Integer.toString(value) );
    return value;
}
    
    
double radians( int angle ) {
    return ( ( Math.PI * angle ) / 180.0 );
}
    
    
double radians( double angle ) {
    return ( ( Math.PI * angle ) / 180.0 );
}


double degrees( double angle ) {
    return ( ( angle * 180.0 ) / Math.PI );
}


void printOut() {
    System.out.println("  --------------------- ");
    System.out.println("Knee Angle: " + kneeAngle );
    System.out.println("Torso Length: " + torsoLength +
		       "  (statedTorso: "+statedTorso+")\n" );
    System.out.println("Thigh = " + thighLength);
    System.out.println("Shin = " + shinLength+"\n");
    System.out.println("Ankle Angle = "+ ankleAngle+" deg   ");
    System.out.println("Knee Angle = "+ kneeAngle+" deg\n     X = "
		       +kneePos.x+"\n     Y = "+kneePos.y +"\n");
    System.out.println("Hip Angle = "+ hipAngle+" deg\n     X = "
		       +hipPos.x+"\n     Y = "+hipPos.y+"\n" );
    System.out.println("Shoulder   X = "
		       +shoulderPos.x+"\n          Y = "+shoulderPos.y +"\n");
}


public void resize (int w, int h) {
    // System.out.println(" resize (int w, int h) is sending " + h );
    adjustHeight( h );
    super.resize( w, h );
}
    
    
public void resize (Dimension d) {
    // System.out.println(" resize (Dimension d) is sending " + d.height );
    adjustHeight( d.height );
    super.resize( d );
}
    
    
public void reshape (int x, int y, int w, int h) {
    if ( h > 0 ) {
	// System.out.println(" reshape is sending " + h );
	adjustHeight( h );
    }
    super.reshape( x,y,w,h );
}
    
    
public synchronized void adjustHeight ( int h ) {
    if ( h > 0 ) 
    	windowHeight = h;
   	try {
   		setVars();
   		update( getGraphics() );
   		getToolkit().sync(); 
   	}  catch( NullPointerException e)  {
	    validate();
	    repaint();
	}
}
    
}
